import isNode from '../../helpers/isNode';
import stringOrIntToMs from '../../helpers/stringOrIntToMs';
import { inMemory } from './handlers/inMemory';
import { inRedis } from './handlers/inRedis';
import LRU from './algorithms/LRU';
import RedisLRU from './algorithms/Redis';
var ALGORITHMS = {
  LRU: 'lru'
};
var DEFAULT_TTL = 500;
var caches = {};

var _cache = function _cache(_temp) {
  var _ref = _temp === void 0 ? {} : _temp,
      algorithm = _ref.algorithm,
      fnName = _ref.fnName,
      instance = _ref.instance,
      original = _ref.original,
      size = _ref.size,
      target = _ref.target,
      ttl = _ref.ttl,
      redis = _ref.redis,
      cacheKeyString = _ref.cacheKeyString;

  var shouldUseRedisCache = redis && isNode;
  var shouldUseMemoryCache = !isNode || !redis;
  var cacheKey = cacheKeyString || target.constructor.name + "::" + fnName;
  var cache = caches[cacheKey];

  if (!cache && shouldUseRedisCache) {
    console.warn("[sui-decorators/cache] You are using redis cache for cacheKey: " + cacheKey + ", your method MUST return a promise");
    cache = new RedisLRU({
      size: size,
      redisConnection: redis,
      namespace: cacheKey,
      ttl: ttl
    });
    caches[cacheKey] = cache;
  }

  if (!cache && shouldUseMemoryCache) {
    redis && console.warn('[sui-decorators/cache] Your redis config will be ignored in client side. Using the default inMemory LRU strategy.');
    cache = new LRU({
      size: size
    });
    caches[cacheKey] = cache;
  }

  return shouldUseRedisCache ? inRedis(target, cache, original, fnName, instance, ttl, cacheKey) : inMemory(target, cache, original, fnName, instance, ttl);
};

var _invalidateCache = function _invalidateCache(_ref2) {
  var original = _ref2.original,
      instance = _ref2.instance,
      cacheKeyString = _ref2.cacheKeyString;
  var cache = caches[cacheKeyString];

  if (cache) {
    cache.clear();
  }

  return function () {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return original.apply(instance, args);
  };
};

export var cache = function cache(_temp2) {
  var _ref3 = _temp2 === void 0 ? {} : _temp2,
      _ref3$ttl = _ref3.ttl,
      ttl = _ref3$ttl === void 0 ? DEFAULT_TTL : _ref3$ttl,
      _ref3$server = _ref3.server,
      server = _ref3$server === void 0 ? false : _ref3$server,
      _ref3$algorithm = _ref3.algorithm,
      algorithm = _ref3$algorithm === void 0 ? ALGORITHMS.LRU : _ref3$algorithm,
      size = _ref3.size,
      redis = _ref3.redis,
      cacheKeyString = _ref3.cacheKeyString;

  var timeToLife = stringOrIntToMs({
    ttl: ttl
  }) || DEFAULT_TTL;
  return function (target, fnName, descriptor) {
    if (typeof window !== 'undefined' && window.__SUI_CACHE_DISABLED__ || typeof global !== 'undefined' && global.__SUI_CACHE_DISABLED__) {
      return descriptor;
    } // if we're on node but the decorator doesn't have the server flag
    // then we ignore the usage of the decorator and thus the cache


    if (isNode && !server) {
      return descriptor;
    }

    if (!isNode && redis) {
      console.warn('[sui-decorators/cache] Your redis config will be ignored in client site. Using the default LRU strategy.');
    }

    var configurable = descriptor.configurable,
        enumerable = descriptor.enumerable,
        writable = descriptor.writable;
    var originalGet = descriptor.get;
    var originalValue = descriptor.value;
    var isGetter = !!originalGet; // https://github.com/jayphelps/core-decorators.js/blob/master/src/autobind.js

    return Object.assign({}, {
      configurable: configurable,
      enumerable: enumerable,
      get: function get() {
        var fn = isGetter ? originalGet.call(this) : originalValue;

        if (this === target) {
          return fn;
        }

        var _fnCached = _cache({
          algorithm: algorithm,
          fnName: fnName,
          instance: this,
          original: fn,
          size: size,
          target: target,
          ttl: timeToLife,
          redis: redis,
          cacheKeyString: cacheKeyString
        });

        Object.defineProperty(this, fnName, {
          configurable: configurable,
          writable: writable,
          enumerable: enumerable,
          value: _fnCached
        });
        return _fnCached;
      },
      set: function set(newValue) {
        Object.defineProperty(this, fnName, {
          configurable: true,
          writable: true,
          enumerable: true,
          value: newValue
        });
        return newValue;
      }
    });
  };
};
export var invalidateCache = function invalidateCache(_temp3) {
  var _ref4 = _temp3 === void 0 ? {} : _temp3,
      cacheKeyString = _ref4.cacheKeyString;

  return function (target, fnName, descriptor) {
    var configurable = descriptor.configurable,
        enumerable = descriptor.enumerable,
        writable = descriptor.writable;
    var originalGet = descriptor.get;
    var originalValue = descriptor.value;
    var isGetter = !!originalGet;
    return Object.assign({}, {
      configurable: configurable,
      enumerable: enumerable,
      get: function get() {
        var fn = isGetter ? originalGet.call(this) : originalValue;

        if (this === target) {
          return fn;
        }

        var _fnCached = _invalidateCache({
          instance: this,
          original: fn,
          cacheKeyString: cacheKeyString
        });

        Object.defineProperty(this, fnName, {
          configurable: configurable,
          writable: writable,
          enumerable: enumerable,
          value: _fnCached
        });
        return _fnCached;
      },
      set: function set(newValue) {
        Object.defineProperty(this, fnName, {
          configurable: true,
          writable: true,
          enumerable: true,
          value: newValue
        });
        return newValue;
      }
    });
  };
};
export default cache;